import great_expectations as gx
from great_expectations.data_context import FileDataContext
import os 
import logging
from typing import Optional, List
from great_expectations.expectations.expectation import ExpectationConfiguration
from great_expectations.core import ExpectationSuite
from feature_pipeline.utils import get_loggers
logs = get_loggers(__name__)
class DataContext:
    """A class to manage Great Expectations data validation context and expectations.
    
    This class handles the creation and management of Great Expectations data context
    and expectation suites for validating energy consumption data.
    """
    
    def __init__(
        self,
        logger: logging.Logger,
        features_list: List[str],
        directory: Optional[str] = None
    ):
        """Initialize the DataContext.
        
        Args:
            logger: Logger instance for logging messages
            features_list: List of feature names to validate
            directory: Optional directory path for Great Expectations context.
                      If None, uses the current working directory
        """
        self.directory = directory 
        self.logger = logger
        self.features_list = features_list
        
        if not os.path.exists(self.directory):
            raise FileNotFoundError(f"Directory not found: {self.directory}")
            
        if not features_list:
            raise ValueError("features_list cannot be empty")

    def create_context_template(self) -> Optional[FileDataContext]:
        """Create a new Great Expectations context template.
        
        Returns:
            FileDataContext if successful, None if creation fails
        """
        try:
            path_to_empty_folder = os.path.join(self.directory, "great_expectations")
            os.makedirs(path_to_empty_folder, exist_ok=True)
            context = FileDataContext._create(project_root_dir=path_to_empty_folder)
            return context
        except Exception as e:
            self.logger.error(f"Failed to create context template: {str(e)}")
            return None

    def build_expectation_suites(self) -> ExpectationSuite:
        """Set up expectation suites for energy consumption data validation.
        
        Returns:
            ExpectationSuite: The created expectation suite
            
        Raises:
            RuntimeError: If context creation fails or expectation setup fails
        """
        try:
            # context = gx.get_context(project_root_dir=self.directory)
            # print(context)
            # if context is None:
            #     context = self.create_context_template()
            #     if context is None:
            #         raise RuntimeError("Failed to create or get context")

            expectation_suite = ExpectationSuite(
            expectation_suite_name="energy_consumption_suite",
            overwrite_existing=True
            )
            
            expectation_configuration = ExpectationConfiguration(
                expectation_type = "expect_The_features_to_macth_oredered_list",
                kwargs = {
                    "column_list":[
                        self.features_list[0],
                        self.features_list[1],
                        self.features_list[2],
                        self.features_list[3]    
                    ]
                },
                meta={
                    "notes": {
                        "format": "markdown",
                        "content": "Validates that the feature columns match the expected order and names"
                    }
                }
            )
            #no null column area
            expectation_configuration_area = ExpectationConfiguration(
                  expectation_type = "expect_column_distinct_values_to_be_in_set",
                  kwargs = {"column":"area",
                            "value_set":(0,1,2)},
            ) 
            #datetime
            expectation_configuration_datetime= ExpectationConfiguration(
                 expectation_type = "expect_column_values_to_not_be_null",
                  kwargs = {"column":"datetime_utc"}
            ) 
            #expect column values to be of type 
            expectation_configuration_4 = ExpectationConfiguration(
                expectation_type="expect_column_values_to_be_of_type",
                kwargs = {"column":"area","type_":"int8"}
            ) 
            expectation_configuration_5 = ExpectationConfiguration(
                 expectation_type="expect_column_distinct_values_to_be_in_set",
            kwargs={
                "column": "consumer_type",
                "value_set": (
                    111,
                    112,
                    119,
                    121,
                    122,
                    123,
                    130,
                    211,
                    212,
                    215,
                    220,
                    310,
                    320,
                    330,
                    340,
                    350,
                    360,
                    370,
                    381,
                    382,
                    390,
                    410,
                    421,
                    422,
                    431,
                    432,
                    433,
                    441,
                    442,
                    443,
                    444,
                    445,
                    446,
                    447,
                    450,
                    461,
                    462,
                    999,
                ),
            },
        )
        
            expectation_configuration_6 = ExpectationConfiguration(
                     expectation_type="expect_column_values_to_be_of_type",
                     kwargs={"column": "consumer_type", "type_": "int32"},
                ) 
            expectation_configuration_7 = ExpectationConfiguration(
                    expectation_type="expect_column_min_to_be_between",
                    kwargs={
                    "column": "energy_consumption",
                    "min_value": 0,
                    "strict_min": False,
                    },
                )
            expectation_configuration_8= ExpectationConfiguration(
                 expectation_type="expect_column_values_to_be_of_type",
                 kwargs={"column": "energy_consumption", "type_": "float64"},
            ) 
            expectation_configuration_9= ExpectationConfiguration(
                 expectation_type="expect_column_values_to_not_be_null",
                 kwargs={"column": "energy_consumption"},
            ) 
            
            
            expectation_suite.add_expectation(expectation_configuration)
            expectation_suite.add_expectation(expectation_configuration_area)
            expectation_suite.add_expectation(expectation_configuration_datetime)
            expectation_suite.add_expectation(expectation_configuration_4)
            expectation_suite.add_expectation(expectation_configuration_5)
            expectation_suite.add_expectation(expectation_configuration_6)
            expectation_suite.add_expectation(expectation_configuration_7)
            expectation_suite.add_expectation(expectation_configuration_8)
            expectation_suite.add_expectation(expectation_configuration_9)
            # context.suites.add(expectation_suite)
            return True
        except Exception as e:
            self.logger.error(f"Failed to setup expectation suites: {str(e)}")
            return  None
        
# if __name__ == "__main__":
#     ex= DataContext(logger=logs,features_list=["datetime_utc","area","consumer_type","energy_consumption"],directory="/Users/bakr/Desktop/personal_projects/energy_forecasting/mlforecasting/great_expectation")
#     a = ex.build_expectation_suites()
#     print(a)