import pandas as pd
import logging
from typing import Optional


class Cleaning:
    def __init__(self, logger: logging.Logger):
        self.logger = logger

    def rename_columns(self, df: pd.DataFrame) -> Optional[pd.DataFrame]:
        try:
            df_cleaned = df.copy()
            df_cleaned.drop(columns=["HourDK"], inplace=True)
            df_cleaned.rename(
                columns={
                    "HourUTC": "datetime_utc",
                    "PriceArea": "area",
                    "ConsumerType_DE35": "consumer_type",
                    "TotalCon": "energy_consumption",
                },
                inplace=True
            )
            return df_cleaned
        except pd.errors.DataError as e:
            self.logger.error(f"An error occurred while renaming the columns: {str(e)}")
            return None

    def cast_columns(self, df: pd.DataFrame) -> Optional[pd.DataFrame]:
        try:
            df_cleaned = df.copy()
            df_cleaned["HourUTC"] = pd.to_datetime(df_cleaned["HourUTC"])
            df_cleaned["PriceArea"] = df_cleaned["PriceArea"].astype(str)
            df_cleaned["ConsumerType_DE35"] = df_cleaned["ConsumerType_DE35"].astype("int32")
            df_cleaned["TotalCon"] = df_cleaned["TotalCon"].astype("float64")
            return df_cleaned
        except pd.errors.DataError as e:
            self.logger.error(f"Error occurred while casting the data types: {str(e)}")
            return None
            
    def drop_duplicates(self, df: pd.DataFrame, subset: Optional[list[str]] = None) -> Optional[pd.DataFrame]:
        try:
            df_cleaned = df.copy()
            df_cleaned.drop_duplicates(subset=subset, inplace=True)
            return df_cleaned
        except Exception as e:
            self.logger.error(f"Error while dropping duplicates: {str(e)}")
            return None
        
    def type_transformation(self,data:pd.DataFrame)->pd.DataFrame:
        tmp_data = data.copy()
        tmp_data["energy_consumption"] = tmp_data["energy_consumption"].astype("float")
        tmp_data["datetime_utc"] = pd.to_datetime(tmp_data["datetime_utc"])
        return tmp_data
    
    def encode_area_columns(self, df: pd.DataFrame) -> Optional[pd.DataFrame]:
        try:
            df_cleaned = df.copy()
            area_mapping = {"DK": 0, "DK1": 1, "DK2": 2}
            df_cleaned["area"] = df_cleaned["area"].map(area_mapping)
            if df_cleaned["area"].isna().any():
                raise ValueError("Unknown area values found in the data")
            df_cleaned["area"] = df_cleaned["area"].astype("int8")
            return df_cleaned
        except Exception as e:
            self.logger.error(f"Error in area encoding: {str(e)}")
            return None

  