import great_expectations as gx
import os 
import logging
from typing import Optional, List
from great_expectations.expectations.expectation import ExpectationConfiguration
from great_expectations import expectations as gxe
import pandas as pd
from datetime import datetime



class Validation:
    """A class to manage Great Expectations data validation context and expectations.
    
    This class handles the creation and management of Great Expectations data context
    and expectation suites for validating energy consumption data.
    """
    
    def __init__(
        self,
        logger: logging.log,
        features_list: List[str] = ["datetime_utc","area","consumer_type","energy_consumption"]

    ):
        """Initialize the DataContext.
        
        Args:
            logger: Logger instance for logging messages
            features_list: List of feature names to validate
            directory: Optional directory path for Great Expectations context.
                      If None, uses the current working directory
        """
        self.logger = logger
        self.features_list = features_list
        if not isinstance(features_list, list):
            self.logger.error("features_list must be a list")
        if not features_list or len(features_list) < 4:
            self.logger.error("features_list must contain at least 4 features")
            
        
        try:
            self.context = gx.get_context(mode="ephemeral")
        except Exception as e:
            self.logger.error(f"Failed to initialize Great Expectations context: {str(e)}")

    def build_expectation_suites(self,data:pd.DataFrame) -> gx.ExpectationSuite:
        """Set up expectation suites for energy consumption data validation.
        
        Returns:
            ExpectationValidationResult: The validation result if successful
            None: If validation fails
        """
        try:
            if not isinstance(data, pd.DataFrame):
                 self.logger.error("data must be a pandas DataFrame")
            # Create or get the expectation suite
            suite_name = "energy_forecasting"
            suite = gx.ExpectationSuite(name=suite_name)
            expectation_suite  = self.context.suites.add(suite)
            
            data_source = self.context.data_sources.add_pandas("pandas")
            data_asset = data_source.add_dataframe_asset(name="pd dataframe asset")
            batch_definition = data_asset.add_batch_definition_whole_dataframe("batch definition")
            batch = batch_definition.get_batch(batch_parameters={"dataframe": data})
          
            
            # expectation_configuration_1 = gxe.ExpectTableColumnsToMatchOrderedList([
            #     self.features_list[0],
            #     self.features_list[1],
            #     self.features_list[2],
            #     self.features_list[3]    
            # ])
            
            # expectation_suite.add_expectation(expectation_configuration_1)
            
        
            expectation_configuration_2 = gxe.ExpectColumnDistinctValuesToBeInSet(
                column="area",
                value_set=(0,1,2)
            )
           
            expectation_suite.add_expectation(expectation_configuration_2)
            
            expectation_configuration_3 = gxe.ExpectColumnValuesToNotBeNull(
                column = "datetime_utc"
            )
            expectation_suite.add_expectation(expectation_configuration_3)
            
            expectation_configuration_4 = gxe.ExpectColumnValuesToBeOfType(
               column="area",
               type_ = "int8"
            )
           
            expectation_suite.add_expectation(expectation_configuration_4)
           
            expectation_configuration_5 = gxe.ExpectColumnDistinctValuesToBeInSet(
                column = "consumer_type",
                value_set = (
                    111,
                    112,
                    119,
                    121,
                    122,
                    123,
                    130,
                    211,
                    212,
                    215,
                    220,
                    310,
                    320,
                    330,
                    340,
                    350,
                    360,
                    370,
                    381,
                    382,
                    390,
                    410,
                    421,
                    422,
                    431,
                    432,
                    433,
                    441,
                    442,
                    443,
                    444,
                    445,
                    446,
                    447,
                    450,
                    461,
                    462,
                    999,
                ),
            )
            
            expectation_suite.add_expectation(expectation_configuration_5)
            
            expectation_configuration_6 = gxe.ExpectColumnValuesToBeOfType(
                column = "consumer_type",
                type_ = "int32"
            )
            expectation_suite.add_expectation(expectation_configuration_6)
            
            expectation_configuration_7 = gxe.ExpectColumnMinToBeBetween(
                 column = "energy_consumption",
                 min_value = 0,
                 strict_min  = False

             )
             
            expectation_configuration_8 = gxe.ExpectColumnValuesToBeOfType(
                column = "energy_consumption",
                type_ = "double"
            )
            expectation_suite.add_expectation(expectation_configuration_8)
            
            expectation_configuration_9 = gxe.ExpectColumnValuesToNotBeNull(
                column = "energy_consumption"
            )
            expectation_suite.add_expectation(expectation_configuration_9)
        
            
            validation_result = batch.validate(expectation_suite)

            if not validation_result.success:
                self.logger.warning("Validation failed: %s", validation_result)
                return None
            self.logger.info("Validation Complete, validation results below")
            self.logger.info("%s", validation_result)
            return validation_result
            
        except Exception as e:
            self.logger.error("Failed to setup expectation suites: %s", str(e))
            return None
        
        

# if __name__ == "__main__":


#     #  "HourUTC": "datetime_utc",
#     #                 "PriceArea": "area",
#     #                 "ConsumerType_DE35": "consumer_type",
#     #                 "TotalCon": "energy_consumption",
#     logger = logging.getLogger(__name__)
#     logging.basicConfig(filename=os.getcwd()+"/outputs/logs.log",encoding="utf-8",level=logging.INFO)
#     data = pd.DataFrame({
#         "datetime_utc": [datetime.now()],  # Note: wrapped in list to create a single row
#         "area": [1],
#         "consumer_type": [1],
#         "energy_consumption": [65]
#     })
#     print(data)
#     feature_list = ["datetime_utc","area","consumer_type","energy_consumption"]
#     ex = Validation(logger,feature_list,data)
#     print(ex.build_expectation_suites())





