import hopsworks
import logging
from typing import Optional
import os
from settings import load_env_var
from utils import Utils 
import wandb
from typing import Tuple
import pandas as pd
from sktime.forecasting.model_selection import temporal_train_test_split
import tempfile
from generating_data import Generate_Data
from databases import connect_to_mongodb , Bulk_insert_into_mongodb
class Data:
    def __init__(self,
                 logger: logging.Logger,
                 version: int,
                 description: Optional[str],
                 project_name: Optional[str],
                 feature_gp_name: str = "energy_forecast"):
        
      
        self.logger = logger
    
        self.feature_gp_name = feature_gp_name
        self.version = version
        self.description = description
        self.project_name = project_name
        self.api_key = os.environ.get("HOPS_WORKS_API")
        self.util = Utils()
        self.api_key_wdb = os.environ.get("WDB_API_KEY")
    def load_data_from_feature_store(self,
         name:Optional[str]=None,
         project_name :Optional[str]=None,
         group:Optional[str]=None,
         job_type:Optional[str]=None,
         entity:Optional[str]=None,
         resume:Optional[str]="allow",
         reinit:Optional[bool]=False,
         run_id:Optional[int]=None,
         add_timestamp_to_name:Optional[bool]=False
        ):
        """"
        Function for loading The data from feature store
        """
        try:
            client_mongo = connect_to_mongodb(self.logger,
                                              os.environ.get("MONGODB_DB_NAME"),
                                              os.environ.get("MONGO_DB_COLLECTION_NAME ")
            )
            # Connect to hopsworks
            try:
                project = hopsworks.login(
                    port=443,
                    project=self.project_name,
                    api_key_value=self.api_key
                )
            except Exception as e:
                self.logger.error(f"Error while connecting to hopsworks: {e}")
                raise  # Re-raise the exception to prevent continuing with invalid connection

            fs = project.get_feature_store(name="driver_score_featurestore")
            
            # Get or create feature group
            try:
                
                fg = fs.get_feature_group(name=self.feature_gp_name, version=self.version)
               
            except:
             
                fg = fs.create_feature_group(
                    name=self.feature_gp_name,
                    version=self.version,
                    description=self.description,
                    primary_key=['datetime_utc', 'area'],
                    online_enabled=True,
                    event_time='datetime_utc'
                )
            
            # Single wandb run for all operations
            with self.util.init_wandb(
                name="loading_dataset",
                project_name=os.environ.get("WDB_PROJECT_NAME"),
                group="datasets",
                job_type="prepare_dataset",
                entity=entity,
                run_id=run_id,
                api_key=self.api_key_wdb
            ) as run:
                # Get feature view and training data
                feature_view = fs.get_feature_view(
                    name="energy_consumption_denmark_view_new",
                    version=self.version
                )
                training_data,_ = Generate_Data.generate_synthetic_time_series()
                
                # training_data, _ = feature_view.get_training_data(
                #     training_dataset_version=1
                # )
                # Log feature view metadata
                fv_metadata = feature_view.to_dict()
                fv_metadata["query"] = fv_metadata["query"].to_string()
                fv_metadata["features"] = [f.name for f in fv_metadata["features"]]
                fv_metadata["link"] = feature_view._feature_view_engine._get_feature_view_url(feature_view)
                fv_metadata["feature_view_version"] = self.version
                fv_metadata["training_dataset_version"] = 1
                artifacts = wandb.Artifact(
                    name="energy_consumption_denmark_feature_view_new",
                    type="feature_view",
                    metadata=fv_metadata
                )
                run.log_artifact(artifacts)
                
                run.finish()

                with self.util.init_wandb(
                    api_key=self.api_key_wdb,
                    name="train_test_split",
                    group="datasets",
                    job_type="prepare_dataset",
                    entity=os.environ.get("WDB_ENTITY_NAME"),
                    run_id=run_id,
                    project_name=os.environ.get("WDB_PROJECT_NAME")
                ) as run:
                    run.use_artifact("energy_consumption_denmark_feature_view_new:latest")
                # Prepare and log split data
                    y_train, y_test, X_train, X_test = self.prepare_data(training_data)
                    splits = {
                        "train": (X_train, y_train),
                        "test": (X_test, y_test)
                    }

                    for split_name, (split_X, split_y) in splits.items():
                       X = tempfile.NamedTemporaryFile(mode='w+', delete=False, suffix='.csv', encoding='utf-8', newline='')
                       Y =  tempfile.NamedTemporaryFile(mode='w+', delete=False, suffix='.csv', encoding='utf-8', newline='')            
                       split_metadata = {
                       "timespan": [
                           split_X["datetime_utc"].min(),
                           split_X["datetime_utc"].max(),
                       ],
                       "dataset_size": len(split_X),
                       "num_areas": len(split_X["area"].unique()),
                       "num_consumer_types": len(split_X["consumer_type"].unique()),
                       "y_features": split_y.columns.tolist(),
                       "X_features": split_X.columns.tolist(),
                       }
                       split_X.to_csv(X.name,index=False)
                       split_y.to_csv(Y.name,index=False)
                       
                       split_artifact = wandb.Artifact(
                           name=f"split_{split_name}",
                           type="split",
                           metadata=split_metadata,
                           )
                       split_artifact.add_file(X.name)
                       split_artifact.add_file(Y.name)
                       run.log_artifact(split_artifact)
    
                    run.finish()
                print(y_train,"heheheh")  
                return y_train, y_test, X_train, X_test
                
        except Exception as e:
            print("error")
            self.logger.error(f"An error occurred in load_data_from_feature_store: {e}")
            return None, None, None, None
    


    def prepare_data(self, data: pd.DataFrame, target: str = "energy_consumption", fh: int = 24):

        try:
            new_data = data.copy()
            new_data["datetime_utc"] = pd.to_datetime(new_data["datetime_utc"])        
            new_data = new_data.set_index(["area", "consumer_type", "datetime_utc"]).sort_index()
            x = new_data.drop(columns=[target])
            y_target = new_data[[target]]    
            y_train, y_test, X_train,X_test = temporal_train_test_split(y=y_target,X=x,fh=fh)
            X_train.index.set_names(["area", "consumer_type", "datetime_utc"],inplace=True)
            X_test.index.set_names(["area", "consumer_type", "datetime_utc"],inplace=True)
            X_train = X_train.reset_index()
            X_test = X_test.reset_index()
            # print(pd.MultiIndex.from_tuples(X_train,names=["area", "consumer_type", "datetime_utc"]))
            return y_train, y_test, X_train,X_test
        except Exception as e:
            self.logger.error(f"Error while processing the data: {e}")
            return None,None,None,None,None



# Working succesfully 
# if __name__ == "__main__":
#     logger = logging.getLogger(__name__)
#     load_env_var("/Users/bakr/Desktop/personal_projects/energy_forecasting/mlforecasting/training_pipeline/")
#     logging.basicConfig(filename=os.getcwd()+"/outputs/logs.log",encoding="utf-8",level=logging.INFO)
#     data  = Data(logger, project_name="driver_score",version=1,description="none")     
    
#     y_train,y_test,X_train,X_test = data.load_data_from_feature_store(     
#         name="test",
#         project_name=os.environ.get("WDB_PROJECT_NAME"),
#         entity="risers",
#         run_id=3,
#         reinit='finish_previous'
#     )
#     print(y_train)
   






    