import logging
import pandas as pd
import numpy as np
from exceptions import ModelTrainingErros
from typing import Tuple,OrderDict,Optional
import hopsworks
import wandb
import os
from settings import load_env_var
from sktime.utils.plotting import plot_series 
from data import Data
from utils import Utils
import json
from models import Models
from sktime.performance_metrics.forecasting import (
    mean_squared_percentage_error,
    mean_absolute_percentage_error,
)
from pathlib import Path

class Train:
    def __init__(self,logger:logging.Logger,api_hopsworks,api_wandb):
        self.logger = logger
        self.api_hopsworks = api_hopsworks
        self.api_wandb = api_wandb
        
        load_env_var("/Users/bakr/Desktop/personal_projects/energy_forecasting/mlforecasting/training_pipeline/")
    
    def from_best_config(self,
                         fh:int=24,
                         feature_view_version:Optional[int]=1,
                         training_dataset_version:Optional[int]=1
                         ):
        #Load The data from feature Store
        self.data_class = Data(self.logger,1,project_name="driver_score")
        y_train,y_test,X_train,X_test = self.data_class.load_data_from_feature_store(
            name="The Test",
            project_name=os.environ.get("WDB_PROJECT_NAME"),
            group="gp1",
            job_type="testing",
            entity=os.environ.get("WDB_ENTITY_NAME"),
            run_id=2
        )
        training_start_datetime = y_train.index.get_level_values("datetime_utc").min()
        training_end_datetime = y_train.index.get_level_values("datetime_utc").max()
        testing_start_datetime = y_test.index.get_level_values("datetime_utc").min()
        testing_end_datetime = y_test.index.get_level_values("datetime_utc").max()
        self.logger.info(
            f"Training model on data from {training_start_datetime} to {training_end_datetime}."
        )
        self.logger.info(
            f"Testing model on data from {testing_start_datetime} to {testing_end_datetime}."
        )
        with Utils.init_wandb(
            api_key=self.api_wandb,
            name = "train_test_split",
            project_name=os.environ.get("WDB_PROJECT_NAME"),
            group="gp1",
            job_type="testing",
            entity=os.environ.get("WDB_ENTITY_NAME"),
            run_id=1
        ) as run:
            run.use_artifact("split_train:latest")
            run.use_artifact("split_test:latest")
            best_config_artifact = run.use_artifact(
                "best_config:latest",
                type="model"
            )
            download_dir = best_config_artifact.download()
            config_path = Path(download_dir)/"best_config.json"
            with open(config_path) as f:
                config = json.load(f)
            run.config.update(config)
            baseline_forecaster = Models.build_base_model(seasonal_periodictivity=24)
            baseline_forecaster = self.train_model(model=baseline_forecaster,
                                                   y_train=y_train,
                                                   X_train=X_train,
                                                   fh=24)
            y_pred,results = self.evaluate(baseline_forecaster,y_test=y_test,X_test=X_test)
            slices = results.pop("slices")
            for k , v in results.items():
                self.logger.info(f"Basekine Test {k}; {v}")
            wandb.log({"test":{"baseline":results}})
            wandb.log({"test.baseline.slices":wandb.Table(dataframe = slices)})
            #Build and train the model 
            best_model = Models.build_model(config)
            best_forecaster= self.train_model(best_model,y_train,X_train,24)
            #evaluate The model 
            y_pred , metrics = self.evaluate(best_forecaster,y_test,X_test)
            slices = metrics.pop("slices")
            for k ,v in metrics.items():
                self.logger.info(f"Model_test {k}: {v}")
            wandb.log({"test":{"model":metrics}})
            wandb.log({"test.model.slices":wandb.Table(dataframe=slices)})
            results = OrderDict({"y_train":y_train,"y_test":y_test,"X_train":X_train,"X_text":X_test})
            self.render(results,prefix="image_test")
            best_forecaster = self.train_model(
            model=best_forecaster,
            y_train=pd.concat([y_train, y_test]).sort_index(),
            X_train=pd.concat([X_train, X_test]).sort_index(),
            fh=fh,
        )
            X_forecast = self.compute_forecast_exogenous_variables(X_test,24)
            y_forecast  = self.forecast(best_forecaster,X_forecast)
            results = OrderDict({"y_train":y_train,"y_test":y_test,"y_forecast":y_forecast})
            self.render(results,prefix="image_forecast")
            save_model_path = OUTPUT_DIR / "best_model.pkl"
        utils.save_model(best_forecaster, save_model_path)
        metadata = {
            "experiment": {
                "fh": fh,
                "feature_view_version": feature_view_version,
                "training_dataset_version": training_dataset_version,
                "training_start_datetime": training_start_datetime.to_timestamp().isoformat(),
                "training_end_datetime": training_end_datetime.to_timestamp().isoformat(),
                "testing_start_datetime": testing_start_datetime.to_timestamp().isoformat(),
                "testing_end_datetime": testing_end_datetime.to_timestamp().isoformat(),
            },
            "results": {"test": metrics},
        }
        artifact = wandb.Artifact(name="best_model", type="model", metadata=metadata)
        artifact.add_file(str(save_model_path))
        run.log_artifact(artifact)

        run.finish()
        artifact.wait()

        model_version = self.attach_best_model_to_feature_store(
            feature_view_version, training_dataset_version, artifact
        )
    
        metadata = {"model_version": model_version}
        utils.save_json(metadata, file_name="train_metadata.json")
    
        return metadata

    def train_model(self,model,y_train:pd.DataFrame,X_train:pd.DataFrame,fh:int):
        try:
           fh = np.arange(fh)+1
           model.fit(y_train,X=X_train,fh=fh)
           return model
        except ModelTrainingErros as e:
            self.logger.error(f"Caught an error : {str(e)}")
            return None
        
    def evaluate(
       self,forecaster,y_test:pd.DataFrame,X_test:pd.DataFrame     
    )->Tuple[pd.DataFrame,dict]:
        y_pred = forecaster.prefdict(X_test)
        results = dict()
        rmspe = mean_squared_percentage_error(y_test, y_pred, squared=False)
        results["RMSPE"] = rmspe
        mape = mean_absolute_percentage_error(y_test, y_pred, symmetric=False)
        results["MAPE"] = mape
        y_test_slices = y_test.groupby(["area","consumer_type"])
        y_pred_slices = y_pred.groupby(["area","consumer_type"])
        slices = pd.DataFrame(columns=["area", "consumer_type", "RMSPE", "MAPE"])
        for y_test_slice, y_pred_slice in zip(y_test_slices, y_pred_slices):
             (area_y_test, consumer_type_y_test), y_test_slice_data = y_test_slice
             (area_y_pred, consumer_type_y_pred), y_pred_slice_data = y_pred_slice
     
             assert (
                 area_y_test == area_y_pred and consumer_type_y_test == consumer_type_y_pred
             ), "Slices are not aligned."
     
             rmspe_slice = mean_squared_percentage_error(
                 y_test_slice_data, y_pred_slice_data, squared=False
             )
             mape_slice = mean_absolute_percentage_error(
                 y_test_slice_data, y_pred_slice_data, symmetric=False
             )
     
             slice_results = pd.DataFrame(
                 {
                     "area": [area_y_test],
                     "consumer_type": [consumer_type_y_test],
                     "RMSPE": [rmspe_slice],
                     "MAPE": [mape_slice],
                 }
             )
             slices = pd.concat([slices, slice_results], ignore_index=True)

        results["slices"] = slices
        return y_pred, results
    # def add_best_model_to_registry(self,best_model_artifact:wandb.Artifact,project:str)->int:
    #     project = hopsworks.login(
    #         api_key_value=self.api_hopsworks,project=os.environ.get("HOPSWORKS_PROJECT_NAME")
    #     )
    #     #Grab The best model from The Artifacts 
    #     best_model_dir = best_model_artifact.download()
    #     best_model = 
    def render(
            timerseries:OrderDict[str,pd.DataFrame],
            prefix:Optional[str]=None,
            delete_from_disk:bool=True
    ):
        grouped_items = OrderDict()
        for split , items in timerseries.items():
            df = items.reset_index(level=[0,1])
            groups = df.groupby(["area","consumer_type"])
            for group_name,split_group_values in groups:
                group_values = grouped_items.get(group_name,{})
                grouped_items[group_name] = {
                f"{split}": split_group_values["energy_consumption"],
                **group_values,
            }
                

        output_dir = os.environ.get("PATH_OUTPUT_DIR") / prefix if prefix else os.environ.get("PATH_OUTPUT_DIR")
        output_dir.mkdir(parents = True,exist_ok=True)
        for group_names , dataframe_items in grouped_items.items():
            fig,ax = plot_series(*dataframe_items.values(),title=group_names,y_labels = dataframe_items.keys())
            fig.suptitle(f"Area: {group_names[0]} - Consumer type: {group_names[1]}")
            #TODO-savings will be to sql 
            #savings to wandb
            if prefix:
                wandb.log({prefix:wandb.Image(fig)})
            else:
                wandb.log(fig)
            if delete_from_disk:
                # TODO-delete from sql 
                pass 


    def compute_forecast_exogenous_variables(X_test: pd.DataFrame, fh: int):

         X_forecast = X_test.copy()
         X_forecast.index.set_levels(
             X_forecast.index.levels[-1] + fh, level=-1, inplace=True
         )

         return X_forecast 

    def forecast(self,forecaster,X_dataset:pd.DataFrame)->list:
        return forecaster.predict(X_dataset)
            
    def attach_best_models_to_feature_sore(
            self,
            model, 
            version:Optional[int]=1,
            name:Optional[str] = "energy_consumption_denmark_view",
            best_model_artifact = wandb.Artifacts  
    ):
        #login to hopsworks   
        project = hopsworks.login(
            port=443,
            project = os.environ.get("HOPSWORKS_PROJECT_NAME"),
            api_key_value = self.api_hopsworks
        )
        #get feature gp
        fs = project.get_feature_store()
        fv = fs.get_feature_view(
            name = name,
            version = version
        )
        fs_tag = {
        "name": "best_model",
        "version": best_model_artifact.version,
        "type": best_model_artifact.type,
        "url": f"https://wandb.ai/{os.environ.get('WDB_ENTITY_NAME')}/{os.environ.get("WDB_PROJECT_NAME")}/artifacts/{best_model_artifact.type}/{best_model_artifact._name}/{best_model_artifact.version}/overview",
        "artifact_name": f"{os.environ.get('WDB_ENTITY_NAME')}/{os.environ.get("WDB_PROJECT_NAME")}/{best_model_artifact.name}",
        }
        fv.add_tag(name="wandb",value = fs_tag)
        best_model_dir = best_model_artifact.download()
        best_model_path = Path(best_model_dir)/"best_model.pk1"
        best_model_metrics = best_model_artifact.metadata["results"]["test"]
        mr = project.get_model_registry()
        py_model = mr.python.create_model("best_model",metrics = best_model_metrics)
        py_model.save(best_model_path)
        return py_model.version

        




        


 
